Index: ugen.c
===================================================================
RCS file: /home/ncvs/src/sys/dev/usb/ugen.c,v
retrieving revision 1.88
diff -u -r1.88 ugen.c
--- ugen.c	15 Aug 2004 23:39:18 -0000	1.88
+++ ugen.c	4 Sep 2004 01:16:23 -0000
@@ -95,8 +95,8 @@
 #define	UGEN_BBSIZE	1024
 
 #define	UGEN_NISOFRAMES	500	/* 0.5 seconds worth */
-#define UGEN_NISOREQS	6	/* number of outstanding xfer requests */
-#define UGEN_NISORFRMS	4	/* number of frames (miliseconds) per req */
+#define UGEN_NISOREQS	128 /* number of outstanding xfer requests */
+#define UGEN_NISORFRMS	8	/* number of frames (miliseconds) per req */
 
 struct ugen_endpoint {
 	struct ugen_softc *sc;
@@ -108,19 +108,28 @@
 	int state;
 #define	UGEN_ASLP	0x02	/* waiting for data */
 #define UGEN_SHORT_OK	0x04	/* short xfers are OK */
+#define UGEN_READ	0x08
+#define UGEN_WRITE	0x10
 	usbd_pipe_handle pipeh;
 	struct clist q;
 	struct selinfo rsel;
+	struct selinfo wsel;
 	u_char *ibuf;		/* start of buffer (circular for isoc) */
 	u_char *fill;		/* location for input (isoc) */
 	u_char *limit;		/* end of circular buffer (isoc) */
 	u_char *cur;		/* current read location (isoc) */
 	u_int32_t timeout;
+	u_int8_t isize;
+	int numfreereq;
+	struct isoreq *curreq;
 	struct isoreq {
 		struct ugen_endpoint *sce;
 		usbd_xfer_handle xfer;
 		void *dmabuf;
 		u_int16_t sizes[UGEN_NISORFRMS];
+		struct isoreq *next;
+		int busy;
+		int frames;
 	} isoreqs[UGEN_NISOREQS];
 };
 
@@ -171,6 +180,8 @@
 		     usbd_status status);
 Static void ugen_isoc_rintr(usbd_xfer_handle xfer, usbd_private_handle addr,
 			    usbd_status status);
+Static void ugen_isoc_wintr(usbd_xfer_handle xfer, usbd_private_handle addr,
+			    usbd_status status);
 Static int ugen_do_read(struct ugen_softc *, int, struct uio *, int);
 Static int ugen_do_write(struct ugen_softc *, int, struct uio *, int);
 Static int ugen_do_ioctl(struct ugen_softc *, int, u_long,
@@ -316,6 +327,7 @@
 	struct ugen_endpoint *sce;
 	u_int8_t niface, nendpt;
 	int ifaceno, endptno, endpt;
+	int nalts, alt;
 	usbd_status err;
 	int dir;
 
@@ -351,23 +363,39 @@
 	for (ifaceno = 0; ifaceno < niface; ifaceno++) {
 		DPRINTFN(1,("ugen_set_config: ifaceno %d\n", ifaceno));
 		err = usbd_device2interface_handle(dev, ifaceno, &iface);
-		if (err)
-			return (err);
-		err = usbd_endpoint_count(iface, &nendpt);
-		if (err)
-			return (err);
-		for (endptno = 0; endptno < nendpt; endptno++) {
-			ed = usbd_interface2endpoint_descriptor(iface,endptno);
-			endpt = ed->bEndpointAddress;
-			dir = UE_GET_DIR(endpt) == UE_DIR_IN ? IN : OUT;
-			sce = &sc->sc_endpoints[UE_GET_ADDR(endpt)][dir];
-			DPRINTFN(1,("ugen_set_config: endptno %d, endpt=0x%02x"
-				    "(%d,%d), sce=%p\n",
-				    endptno, endpt, UE_GET_ADDR(endpt),
-				    UE_GET_DIR(endpt), sce));
-			sce->sc = sc;
-			sce->edesc = ed;
-			sce->iface = iface;
+		if (err) {
+			printf("ugen usbd_device2interface_handle failed %d\n", ifaceno);
+			continue;
+		}
+		nalts = usbd_get_no_alts(usbd_get_config_descriptor(dev), ifaceno);
+		printf("ugen nalts %d\n", nalts);
+		for(alt = 0; alt<nalts+1; alt++) {
+			err = usbd_set_interface(iface, alt);
+			if (err) {
+				printf("ugen nalts %d %d failed #1\n", ifaceno, alt);
+				DPRINTFN(1,("ugen_set_config: usbd_set_interface error\n"));
+				continue;
+			}
+			err = usbd_endpoint_count(iface, &nendpt);
+			if (err) {
+				printf("ugen nalts %d %d failed #2\n", ifaceno, alt);
+				DPRINTFN(1,("ugen_set_config: endpoint_count error\n"));
+				continue;
+			}
+			for (endptno = 0; endptno < nendpt; endptno++) {
+				ed = usbd_interface2endpoint_descriptor(iface,endptno);
+				endpt = ed->bEndpointAddress;
+				dir = UE_GET_DIR(endpt) == UE_DIR_IN ? IN : OUT;
+				sce = &sc->sc_endpoints[UE_GET_ADDR(endpt)][dir];
+				DPRINTFN(1,("ugen_set_config int %d alt %d: endptno %d, endpt=0x%02x"
+							"(%d,%d), sce=%p\n",
+							ifaceno, alt,
+							endptno, endpt, UE_GET_ADDR(endpt),
+							UE_GET_DIR(endpt), sce));
+				sce->sc = sc;
+				sce->edesc = ed;
+				sce->iface = iface;
+			}
 		}
 	}
 
@@ -413,8 +441,14 @@
 	for (dir = OUT; dir <= IN; dir++) {
 		if (flag & (dir == OUT ? FWRITE : FREAD)) {
 			sce = &sc->sc_endpoints[endpt][dir];
-			if (sce == 0 || sce->edesc == 0)
+			if (sce == 0 || sce->edesc == 0) {
+				DPRINTFN(1, ("ugen failed dir %s #3 sce == 0 = %s\n",dir == OUT?"out":"in", sce ==0 ?"true":"false"));
+				if (dir == IN && flag & FWRITE) {
+					flag &= ~FREAD;
+					continue;
+				}
 				return (ENXIO);
+			}
 		}
 	}
 
@@ -465,25 +499,31 @@
 				return (EIO);
 			break;
 		case UE_ISOCHRONOUS:
-			if (dir == OUT)
-				return (EINVAL);
-			isize = UGETW(edesc->wMaxPacketSize);
-			if (isize == 0)	/* shouldn't happen */
-				return (EINVAL);
-			sce->ibuf = malloc(isize * UGEN_NISOFRAMES,
-				M_USBDEV, M_WAITOK);
-			sce->cur = sce->fill = sce->ibuf;
-			sce->limit = sce->ibuf + isize * UGEN_NISOFRAMES;
-			DPRINTFN(5, ("ugenopen: isoc endpt=%d, isize=%d\n",
-				     endpt, isize));
+			sce->isize = isize = UGETW(edesc->wMaxPacketSize);
+			DPRINTFN(1, ("ugenopen: isoc endpt=%d isize=%d\n", endpt, isize));
 			err = usbd_open_pipe(sce->iface,
-				  edesc->bEndpointAddress, 0, &sce->pipeh);
+					edesc->bEndpointAddress, 0, &sce->pipeh);
 			if (err) {
-				free(sce->ibuf, M_USBDEV);
+				printf("ugen iso %s open usbd_open_pipe %s\n", dir == IN?"in":"out", usbd_errstr(err));
+				if (dir == IN && (sce->state & UGEN_WRITE) )
+					continue;
 				return (EIO);
 			}
+			if (isize == 0) {	/* shouldn't happen */
+				DPRINTFN(1, ("ugenopen: isoc endpt=%d failed: isize=0\n", endpt));
+				return (EINVAL);
+			}
+			if (dir == IN) {
+				sce->ibuf = malloc(isize * UGEN_NISOFRAMES,
+						   M_USBDEV, M_WAITOK);
+				sce->cur = sce->fill = sce->ibuf;
+				sce->limit = sce->ibuf + isize * UGEN_NISOFRAMES;
+			}
+			sce->curreq = sce->isoreqs;
+			sce->numfreereq = UGEN_NISOREQS;
 			for(i = 0; i < UGEN_NISOREQS; ++i) {
 				sce->isoreqs[i].sce = sce;
+				sce->isoreqs[i].frames = 0;
 				xfer = usbd_alloc_xfer(sc->sc_udev);
 				if (xfer == 0)
 					goto bad;
@@ -497,14 +537,20 @@
 				sce->isoreqs[i].dmabuf = buf;
 				for(j = 0; j < UGEN_NISORFRMS; ++j)
 					sce->isoreqs[i].sizes[j] = isize;
-				usbd_setup_isoc_xfer
-					(xfer, sce->pipeh, &sce->isoreqs[i],
-					 sce->isoreqs[i].sizes,
-					 UGEN_NISORFRMS, USBD_NO_COPY,
-					 ugen_isoc_rintr);
-				(void)usbd_transfer(xfer);
+				sce->isoreqs[i].next = sce->isoreqs + (i==0? UGEN_NISORFRMS-1 : i-1);
+				sce->isoreqs[i].busy = 0;
+				if (dir == IN) {
+					usbd_setup_isoc_xfer
+						(xfer, sce->pipeh, &sce->isoreqs[i],
+						 sce->isoreqs[i].sizes,
+						 UGEN_NISORFRMS, USBD_NO_COPY,
+						 ugen_isoc_rintr);
+					(void)usbd_transfer(xfer);
+				}
+			sce->state |= dir==IN?UGEN_READ:UGEN_WRITE;
+	
 			}
-			DPRINTFN(5, ("ugenopen: isoc open done\n"));
+			DPRINTFN(5, ("ugenopen: isoc %s open done\n",dir == IN?"IN":"OUT"));
 			break;
 		bad:
 			while (--i >= 0) /* implicit buffer free */
@@ -610,6 +656,11 @@
 	if (sce == NULL)
 		return (EINVAL);
 
+	if (!(sce->state & UGEN_READ)) {
+		DPRINTFN(1, ("ugenread not in read mode\n"));
+		return (ENXIO);
+	}
+
 	if (sce->edesc == NULL) {
 		printf("ugenread: no edesc\n");
 		return (EIO);
@@ -748,7 +799,9 @@
 {
 	struct ugen_endpoint *sce = &sc->sc_endpoints[endpt][OUT];
 	u_int32_t n;
+	int s;
 	int error = 0;
+	int sofar;
 	char buf[UGEN_BBSIZE];
 	usbd_xfer_handle xfer;
 	usbd_status err;
@@ -758,22 +811,83 @@
 	if (sc->sc_dying)
 		return (EIO);
 
-	if (endpt == USB_CONTROL_ENDPOINT)
+	if (endpt == USB_CONTROL_ENDPOINT) {
+		DPRINTFN(1, ("ugenwrite: endpoint not allowed\n"));
 		return (ENODEV);
+	}
 
-	if (sce == NULL)
+	if (sce == NULL) {
+		DPRINTFN(1, ("ugenwrite: no sce\n"));
 		return (EINVAL);
+	}
+	if (!(sce->state&UGEN_WRITE) ) {
+		DPRINTFN(1, ("ugenwrite: not in write mode\n"));
+		return (EINVAL);
+	}
 
 	if (sce->edesc == NULL) {
-		printf("ugenwrite: no edesc\n");
+		DPRINTFN(1, ("ugenwrite: no edesc sce=%p\n",sce));
 		return (EIO);
 	}
 	if (sce->pipeh == NULL) {
-		printf("ugenwrite: no pipe\n");
+		DPRINTFN(1, ("ugenwrite: no pipe sce=%p\n",sce));
 		return (EIO);
 	}
 
 	switch (sce->edesc->bmAttributes & UE_XFERTYPE) {
+	case UE_ISOCHRONOUS:
+		sofar = 0;
+		s = splusb();
+		while (uio->uio_resid > 0 && !error) {
+			while ( sofar > (UGEN_NISOREQS>>4) ||
+					sce->curreq->busy) {
+				sofar = 0;
+				if (flag & IO_NDELAY) {
+					splx(s);
+					return (EWOULDBLOCK);
+				}
+				sce->state |= UGEN_ASLP;
+				DPRINTFN(5, ("ugenwrite: sleep on %p\n", sce));
+				error = tsleep(sce, PZERO | PCATCH, "ugenws", 0);
+				DPRINTFN(5, ("ugenwrte: woke, error=%d\n", error));
+				if (sc->sc_dying)
+					error = EIO;
+				if (error) {
+					sce->state &= ~UGEN_ASLP;
+					break;
+				}
+			}
+			/*
+			 * sce->curreq is ready to go
+			 */
+			while(!error && sce->curreq->frames<UGEN_NISORFRMS && uio->uio_resid>0) {
+				n = min(sce->isize, uio->uio_resid);
+
+				DPRINTFN(5, ("ugenwrte isoc got %d chars\n", n));
+
+				/* Copy the data from the user process. */
+				error = uiomove((u_char *)sce->curreq->dmabuf+(sce->curreq->frames*sce->isize), n, uio);
+				sce->curreq->sizes[sce->curreq->frames]=n;
+				if (error)
+					break;
+				sce->curreq->frames++;
+			}
+			if (!error && (sce->curreq->frames==UGEN_NISORFRMS) ) {
+				sce->curreq->busy=1;
+				sce->numfreereq--;
+				usbd_setup_isoc_xfer(sce->curreq->xfer,
+						sce->pipeh, sce->curreq,
+						sce->curreq->sizes, sce->curreq->frames,
+						USBD_NO_COPY, ugen_isoc_wintr);
+				(void)usbd_transfer(sce->curreq->xfer);
+				sce->curreq->frames=0;
+				sce->curreq=sce->curreq->next;
+				//sofar++;
+			}
+		}
+		splx(s);
+		break;
+
 	case UE_BULK:
 		xfer = usbd_alloc_xfer(sc->sc_udev);
 		if (xfer == 0)
@@ -950,6 +1064,33 @@
 }
 
 Static void
+ugen_isoc_wintr(usbd_xfer_handle xfer, usbd_private_handle addr,
+		usbd_status status)
+{
+	struct isoreq *req = addr;
+	struct ugen_endpoint *sce = req->sce;
+	u_int32_t count;
+
+	/* Return if we are aborting. */
+	if (status == USBD_CANCELLED)
+		return;
+
+	usbd_get_xfer_status(xfer, NULL, NULL, &count, NULL);
+	DPRINTFN(5,("ugen_isoc_wintr: xfer %d, count=%d\n",
+		    (int)(req - sce->isoreqs),
+		    count));
+
+	//isize = UGETW(sce->edesc->wMaxPacketSize);
+	req->busy = 0;
+	sce->numfreereq++;
+	if (sce->numfreereq > (UGEN_NISOREQS/2) && sce->state & UGEN_ASLP) {
+		sce->state &= ~UGEN_ASLP;
+		DPRINTFN(5, ("ugen_isoc_wintr: waking %p\n", sce));
+		wakeup(sce);
+	}
+	selwakeuppri(&sce->wsel, PZERO);
+}
+Static void
 ugen_isoc_rintr(usbd_xfer_handle xfer, usbd_private_handle addr,
 		usbd_status status)
 {
@@ -1449,6 +1590,12 @@
 			else
 				selrecord(p, &sce->rsel);
 		}
+		if (events & (POLLOUT | POLLWRNORM)) {
+			if (sce->cur != sce->fill)
+				revents |= events & (POLLOUT | POLLWRNORM);
+			else
+				selrecord(p, &sce->wsel);
+		}
 		break;
 	case UE_BULK:
 		/*
